#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#define extern static
#  include <asm/io.h>
#undef extern
#include <sys/types.h>

/*****************************************************/
#define USE_LPT 1
#define USE_ALSA 1

#include "mtrx.h"
#define MAIN
#include "mrx.h"
#if USE_ALSA
#include "mrx_alsa.h"
#endif
#if USE_LPT
#include "mrx_lpt.h"
#endif
#undef MAIN

/*****************************************************/
char *usage = 
#if USE_ALSA
              "Usage (Soundcard with ALSA driver):\n"
              "mrx -h0 -f FREQUENCY -l DETECT_LEVEL\n"
#endif
#if USE_LPT
              "Usage (LPT hardware input):\n" 
              "mrx -h1 -p PORTNR -b PORTBIT -i INVERTBIT -f FREQUENCY\n"
#endif
              "";

static int Hardware=HARDWARE;   /* 0=Alsa, 1=LPT */
static int ControlPipe=0;       /* 0=no, 1=yes */
static int Threshold = START_THRESHOLD; /* Speed at start time */
static int Fdes;                /* Control Pipe */

/*****************************************************/

int main_init(int argc, char *argv[]);
int opencontrolpipe(const char *fifoname);
char getcw(char *m2a);
  int testdit(void);
     void getdit(int *poldbit, int *plength);
        void getcommand(int fdes);
        int getstat(void);
        long getduration(struct timeval *pt2, struct timeval *pt1);
  int speedcontrol(int dit, int threshold, int length);
void putascii(char ch);
void putinfo(char itype);

/*****************************************************/

int main(int argc, char *argv[])
{
   char m2a[MMAX];
   printf("THIS IS MRX " Version "\n"); fflush(stdout);
   m2ainit(m2a,MMAX);
   if(!main_init(argc,argv))
   {
      fprintf(stderr, "%s", usage);
      return 1;
   }

   if(ControlPipe)
   {
      Fdes=opencontrolpipe(FIFONAME);
      if(Fdes==-1)
      {
         fprintf(stderr, "opencontrolpipe(): error!\n", usage);
         return 3;
      }
      putinfo('S');
      putinfo('L');
      putinfo('F');
   }
   
   switch(Hardware)
   {
#if USE_ALSA
   case 0: init_alsa(); break;
#endif
#if USE_LPT
   case 1: init_lpt();   break;
#endif
   default: fprintf(stderr,"unknown hardware type: exit!\n"); return 3;
   }
   
   while(1)
   {
      char ch;
      ch=getcw(m2a);
      putascii(ch);
   }
   
   switch(Hardware)
   {
#if USE_ALSA
   case 0: exit_alsa(); break;
#endif
#if USE_LPT
   case 1: exit_lpt();  break;
#endif
   default: fprintf(stderr,"unknown hardware type: exit!\n"); return 2;
   }

   return 0;

}/* main() */

/********************************************************************

NAME
	main_init()

SYNOPSIS
	main_init(int argc, char *argv[])

DESCRIPTION
        sets global variables

ARGUMENTS
        int argc, char *argv[]: like main
        
RETURN VALUE
         0 : success
        -1 : unknown option

NOTE
        uses some global variables:
        Hardware from mrx.h
        InPort, InBit, InInvert from mrx_hw.h
        Frequency, DetectLevel from mrx_alsa.h
        
SEE ALSO
        ---
Date       Changes
---------------------------------------------------------
xx.xx.xxxx New
22.10.2008 command line options instead of env. variables

*********************************************************************/

int main_init(int argc, char *argv[])
{
   const char options[]="ch:"
#if USE_LPT
                         "p:b:i:"
#endif
#if USE_ALSA
                         "l:f:"
#endif
                         ;
   int rc;	
   
   while((rc=getopt(argc, argv, options))!=EOF)
   {
      switch(rc)
      {
         case 'c': ControlPipe=1; break;
         case 'h': Hardware=atol(optarg);break;
#if USE_LPT
         case 'p': InPort=atol(optarg);  break;
         case 'b': InBit=atol(optarg);   break;
         case 'i': InInvert=atol(optarg);break;
#endif
#if USE_ALSA
         case 'l': DetectLevel=atof(optarg);break;
         case 'f': Frequency=atof(optarg);  break;
#endif
         default:
		fprintf(stderr,"unknown option(\"%c\")!\n", rc);
		return -1;
	    }/* switch */
	}/* for options */
   
}/* main_init() */

/*******************************************************************/

int opencontrolpipe(const char *fifoname)
{
   int fdes;
   int rc;
   rc=mknod(fifoname, S_IFIFO|0600, 0);
   if(rc!=0&&errno!=EEXIST)
   {
      fprintf(stderr,"%s,%d: mknod(%s): %s!\n",
      __FILE__, __LINE__, fifoname, strerror(errno));
      return -1;
   }
   
   fdes=open(fifoname, O_RDONLY|O_NONBLOCK);
   if(fdes==-1)
   {
      fprintf(stderr,"%s,%d: open(%s): %s!\n",
      __FILE__, __LINE__, fifoname, strerror(errno));
      return -1;
   }
   return fdes;
      
}/* opencontrolpipe() */

/*******************************************************************/

void getcommand(int fdes)
{
   int rc;
   char s[2];
   rc=read(fdes,s,1);
   if(rc==1)
   {
      switch(s[0])
      {
      case 's': Threshold+=Threshold/10; putinfo('S'); break;
      case 'S': Threshold-=Threshold/10; putinfo('S'); break;
#if USE_ALSA
      case 'l': DetectLevel*=.707; putinfo('L'); break;
      case 'L': DetectLevel*=1.41; putinfo('L'); break;
      case 'f': Frequency-=50.0;   putinfo('F'); break;
      case 'F': Frequency+=50.0;   putinfo('F'); break;
#endif 
      default: break;
      }
      fflush(stdout);
   }
   else if(rc==-1&&errno!=EAGAIN)
   {
      fprintf(stderr,"read():%s!\n", strerror(errno));
   }
   
}/* getcommand() */

/********************************************************************

NAME
	getcw()

SYNOPSIS
	getcw(void)

DESCRIPTION
        calculates last received char out of dits and dahs

ARGUMENTS
        char m2a[]: translation table of characters
        
RETURN VALUE
        char: received char
        
NOTE
        static variable PbwRec

SEE ALSO
        getdit()
        speedcontrol()
        
Date       Changes
------------------
xx.xx.xxxx New

*********************************************************************/

char getcw(char *m2a)
{
   int dit;
   int codenr;
   static int PbwRec=0;
				/* First put out the long pause from the last */
                                /* time as a SPC */
   if(PbwRec)
   {
      PbwRec=0;
      return(SPCSYMBOL);
   }
				/* built the codenr out of the dits and dahs */
   codenr = 1;
   do{
      dit = testdit();
      if(dit==DIT)
      {
         codenr *= 2;
         ++codenr;
      }
      else if(dit==DAH)
      {
         codenr *= 2;
      }
   }
   while((dit!=PAUSEBETWEENCHAR) && (dit!=PAUSEBETWEENWORDS) && (codenr<=32767));

                               	/* This time the char, next time the SPC  */
				/* So it has to be stored in a static Var.*/
   PbwRec = (dit==PAUSEBETWEENWORDS);

				/* A Morse Error ........ can be 8 Points or more */
   switch(codenr)
   {
      case 2047:
      case 4095:
      case 8191:
      case 16383:
      case 32767:
         codenr = 255;
         break;
      case 5734:
         codenr = 730;
         break;
      default:
         break;
   }

#if debug4
   fprintf(stderr,"got codenr: %d (%c)\n", codenr, m2a[codenr]);
#endif
				/* can this received codenr be found in the m2a table? */
   if(codenr>=MMAX || codenr<0)
   {
	return(ERRSYMBOL);
   }
   return(m2a[codenr]);

}/* getcw() */

/********************************************************************

NAME
	testdit()

SYNOPSIS
	testdit(void)

DESCRIPTION
        returns last received part of actual char

ARGUMENTS
        ---
        
RETURN VALUE
        int: part of char, can be DIT, DAH, PAUSEWITHINCHAR,
             PAUSEBETWEENCHAR or PAUSEBETWEENWORDS
NOTE
        static variable Threshold keeps cw speed (length between
        dit length and dah length).

SEE ALSO
        getdit()
        speedcontrol()
        
Date       Changes
------------------
xx.xx.xxxx New

*********************************************************************/

int testdit(void)
{
   int dit;			/* DIT or DAH or Pause or..: output of this function */
   int status;			/* 0- or 1-Status on the input line: input of this function */
   int length;			/* Length of 0- or 1-Status: input of this function */
   int rc;

   getdit(&status, &length);
					/* DIT or DAH ? */
   if(length>Threshold)
   {
      if(status)			/* Long Tone	*/
      {
         dit=DAH;
      } 
      else 				/* Long Pause	*/
      {
#define TEST20081015 0
#if TEST20081015
         if(length>(5*Threshold))
#else
         if(length>(2*Threshold))
#endif
         {
            dit=PAUSEBETWEENWORDS;
         } 
         else				/* Semi-Long Pause */
         {
            dit=PAUSEBETWEENCHAR;
         }
      }
   } 
   else 				/* Short Tone	*/
   {
      if(status)
      {
         dit = DIT;
      } 
      else 				/* Short Pause	*/
      {
         dit = PAUSEWITHINCHAR;
      }
   }

#if debug3
   {
      char d[][4]=DITOUT;
      fprintf(stderr,"%s",d[dit]);
   }
#endif
   
   Threshold = speedcontrol(dit,rc=Threshold,length);
   if(Threshold!=rc && ControlPipe) putinfo('S');
   return dit;
   
}/* testdit() */

/********************************************************************

NAME
	speedcontrol()

SYNOPSIS
	speedcontrol(dit, threshold, length)

DESCRIPTION
        tries to estimate the actual cw speed

ARGUMENTS
        int dit: type of actual part of actual char
        int threshold: actual threshold between dits and dahs
        int length: length of actual part of actual char
        
RETURN VALUE
        new threshold length in milliseconds between dits and dahs

NOTE
        function needs further improvement
        
SEE ALSO
        --- 
Date       Changes
------------------
xx.xx.xxxx New

*********************************************************************/

int speedcontrol(int dit, int threshold, int length)
{
   switch(dit)
   {
      case DIT: 
      {
         int new_value;
         
         new_value=2*length;     
                             /* prevent extreme changes */
         if(new_value<=MAX_CORR*threshold 
         && threshold<=MAX_CORR*new_value)
         {
            threshold += (new_value-threshold)*DIT_ADJ/100;
         }
#        if debug2
	    fprintf(stderr,"new threshold (from dit): %dms\n",threshold);
#        endif 
      }
      break;
      case DAH:
      {
         int new_value;
         new_value=2*length/3;
                               /* prevent extreme changes */
         if(new_value<=MAX_CORR*threshold
         && threshold<=MAX_CORR*new_value)
         {
            threshold += (new_value-threshold)*DAH_ADJ/100;
         }
#        if debug2
	    fprintf(stderr,"new threshold (from dah): %dms\n",threshold);
#        endif
      }
      break;
      case PAUSEBETWEENWORDS:
      case PAUSEBETWEENCHAR:
      case PAUSEWITHINCHAR:
      default: break;
   }
   return threshold;

}/* speedcontrol() */

/********************************************************************

NAME
	putascii()

SYNOPSIS
	putascii(ch)

DESCRIPTION
        Any lower character is printed directly,
        other alphanumeric characters are translated.
        
ARGUMENTS
        char ch: input char
       
RETURN VALUE
       void

SEE ALSO
       --- 
Date       Changes
------------------
xx.xx.xxxx New
24.02.2007 i18n

*********************************************************************/

void putascii(char ch)
{
   switch(ch)
   {
   case '+':/*  53 .-.-.       ar */ printf("%s", "*ar/Spruchende)*");	break;
   case 'A':/*  42 -.-.-       ka */ printf("%s", "*ka/Spruchanfang*");	break;
   case 'B':/* 122 ...-.-      sk */ printf("%s", "*sk/Betriebsende*");	break;
   case 'C':/*  16 ----           */ printf("%s", "ch");		break;
   case 'D':/* 731 -..-..-..  ddd */ printf("%s", "*EMERGENCY*");	break;
   case 'E':/*  38 --..-          */ printf("%s", "*epsilon*");		break;
   case 'G':/* 347 -.-..-..    cl */ printf("%s", "*cl/closed*");	break;
   case 'K':/* 186 -...-.-     bk */ printf("%s", "*bk/break*");	break;
   case 'I':/* 127 .......        */ printf("%s", "*error*");		break;
   case 'S':/* 967 ...---...      */ printf("%s", "* S O S *");		break;
   case 'T':/*  54 .-..-          */ printf("%s", "*+*");		break;
   case 'V':/*  61 ...-.      ve  */ printf("%s", "*verstanden*");	break;
   case 'W':/*  55 .-...      eb  */ printf("%s", "*eb/wait*");		break;
   case 'X':/* 730--.--..--..-xxx */ printf("%s", "*DRINGLICH*");	break;
   default:                          printf("%c", ch); 
   } 
   fflush(stdout);
   return;

}/* putascii() */

/********************************************************************

NAME
	getdit()

SYNOPSIS
	getdit(poldbit, plength)

DESCRIPTION

ARGUMENTS
       int *poldbit
       int *plength
       
RETURN VALUE
       void

NOTE
       getdit() stores the old value of line status and time
       within two static variables.

SEE ALSO
       getstat();
       getduration();
        
Date       Changes
------------------
xx.xx.xxxx New

*********************************************************************/

void getdit(int *poldbit, int *plength)
{
   int            newbit;
   struct timeval newtime;

   static int            FirstTime = 1;
   static int            OldBit;
   static struct timeval OldTime;
   
   if(FirstTime)
   {
      FirstTime = 0;
      OldBit    = 0;
      gettimeofday(&OldTime,NULL);
   }
				/* ask the hardware: does status-bit change? */
   do
   {
      newbit=getstat();
      gettimeofday(&newtime,NULL);	
      if(ControlPipe)      
      {
         getcommand(Fdes);
      }
      usleep(1000l);
   }/* do */
   while(OldBit== newbit);
   
#if debug1
   fprintf(stderr,"new status: >%d<\n",[HL]newbit);
#endif
					/* set output parameters */
					/* calculate *plength */
   *plength = getduration(&newtime, &OldTime);
   if(*plength<=0)
   {
      fprintf(stderr,"Error: Status duration= %dms!\n", *plength);
   }
   *poldbit = OldBit;

   OldBit  = newbit;
   OldTime = newtime;
   return;
   
}/* getdit() */

/********************************************************************

NAME
	getstat()

SYNOPSIS
	int getstat(void);

DESCRIPTION
        Returns actual status of line (0 or 1).

ARGUMENTS
	---

RETURN VALUE + ERRORS
        0,1: Return Status.

GLOBAL VARIABLES
        Hardware

SEE ALSO
        ---
        
Date       Changes
------------------
24.02.2007 New

*********************************************************************/

int getstat(void)
{
   int rc;
   switch(Hardware)
   {
   case 0: rc=getstat_alsa(); break;
   case 1: rc=getstat_lpt(); break;
   default: rc=0;
   }
   return rc;
   
}/* getstat() */

/********************************************************************

NAME
	getduration()

SYNOPSIS
        #include <sys/time.h>
        #include <time.h>
	long getduration(pt2, pt1)

DESCRIPTION
	Calculates the duration between pt1 and pt2.

ARGUMENTS
	struct timeval *pt2, struct timeval *pt1: 
	pt2 is later than pt1

RETURN VALUE + ERRORS
        0L: more than LONG_MAX milliseconds or timer overflow
        else: duration between pt1 and pt2, if pt2 is later than pt1

SEE ALSO
        ---
Date       Changes
------------------
xx.xx.xxxx New

*********************************************************************/

long getduration(struct timeval *pt2, struct timeval *pt1)
{
   long d;
   d  = (pt2->tv_sec  - pt1->tv_sec )*1000;	
   d += (pt2->tv_usec - pt1->tv_usec)/1000;
   if(d<0)
   {
      return 0L; /* more than LONG_MAX milliseconds or timer overflow */
   }
   return d;
   
}/* getduration() */

/********************************************************************/

void putinfo(char itype)
{
   switch(itype)
   {
   case 'S': printf("[S%i]", 2*CONVERT_FWPM_TDIT/Threshold); break;
#if USE_ALSA
   case 'L': printf("[L%5.1le]", DetectLevel); break;
   case 'F': printf("[F%.0lf]", Frequency); break;
#endif
   default: break;                                    
   }
   fflush(stdout);
   
}/* putinfo() */
