#define __USE_BSD
#include <math.h>
#include "mtx_alsa.h"

/********************************************************************

NAME
	init_talsa()

SYNOPSIS
        #include "mtx_alsa.h"
	int init_talsa(void);

DESCRIPTION
        ...
ARGUMENTS
	---

RETURN VALUE + ERRORS
        0: ok
        1: error

NOTES
       needs libasound2-dev (lib-alsa-sound)
SEE ALSO
       --- 
Date       Changes
------------------
25.02.2007 under construction

*********************************************************************/

int init_talsa(void)
{
	int err;
	int drate=T_DRATE;
	snd_pcm_hw_params_t *hw_params;

	if ((err = snd_pcm_open (&playback_handle, "default", SND_PCM_STREAM_PLAYBACK, 0)) < 0) 
	{
		fprintf (stderr, "cannot open audio device for write\n", snd_strerror (err));
		return -1;
	}
	if ((err = snd_pcm_hw_params_malloc (&hw_params)) < 0) 
	{
		fprintf (stderr, "cannot allocate hardware parameter structure (%s)\n", snd_strerror (err));
		return -2;
	}
	if ((err = snd_pcm_hw_params_any (playback_handle, hw_params)) < 0) 
	{
		fprintf (stderr, "cannot initialize hardware parameter structure (%s)\n", snd_strerror (err));
		return -3;
	}
	if ((err = snd_pcm_hw_params_set_access (playback_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) 
	{
		fprintf (stderr, "cannot set access type (%s)\n", snd_strerror (err));
		return -4;
	}
	if ((err = snd_pcm_hw_params_set_format (playback_handle, hw_params, SND_PCM_FORMAT_S16_LE)) < 0)
	{
		fprintf (stderr, "cannot set sample format (%s)\n", snd_strerror (err));
		return -5;
	}
	if ((err = snd_pcm_hw_params_set_rate_near (playback_handle, hw_params, &drate, 0)) < 0) 
	{
		fprintf (stderr, "cannot set sample rate (%s)\n", snd_strerror (err));
		return -6;
	}
	if ((err = snd_pcm_hw_params_set_channels (playback_handle, hw_params, 1)) < 0) 
	{
		fprintf (stderr, "cannot set channel count (%s)\n", snd_strerror (err));
		return -7;
	}
	if ((err = snd_pcm_hw_params (playback_handle, hw_params)) < 0) 
	{
		fprintf (stderr, "cannot set parameters (%s)\n", snd_strerror (err));
		return -8;
	}
	snd_pcm_hw_params_free (hw_params);
	if ((err = snd_pcm_prepare (playback_handle)) < 0) 
	{
		fprintf (stderr, "cannot prepare audio interface for use (%s)\n", snd_strerror (err));
		return -9;
	}
	return 0;
   
}/* init_talsa() */

/********************************************************************

NAME
	putstat_talsa()

SYNOPSIS
        #include "mtx_alsa.h"
	void putstat_talsa(level, duration);

DESCRIPTION
        if level!=0: output of a sine tone
        else: no output
        
ARGUMENTS
	int level
        int duration: duration in milliseconds
        
RETURN VALUE + ERRORS
        void 
NOTES
       needs libasound2-dev (lib-alsa-sound)
       global var: Frequency
       
SEE ALSO
       --- 
Date       Changes
------------------
25.02.2007 under construction

*********************************************************************/
#define DEBUG 0
void putstat_talsa(int level, int duration)
{
   long dur_num_of_frames;
   static int16_t buf[T_NUM_OF_FRAMES];
   static int firsttime=1;
   if(firsttime)
   {
      int i;
      firsttime=0;
      for(i=0; i<T_NUM_OF_FRAMES; ++i)
      {
         buf[i]=(int16_t)(T_LEVEL/100.0*32768.0*sin(2*M_PI*i*T_FREQUENCY/T_DRATE));
  #if DEBUG
                printf(">%i<", buf[i]);
  #endif
      }/* for */
      printf("\n");
   }/* if */
   
   dur_num_of_frames=duration*T_DRATE/1000;
   while(dur_num_of_frames>0)
   {
      int written;
      int to_write;
      
      if(dur_num_of_frames>=T_NUM_OF_FRAMES)
         to_write=T_NUM_OF_FRAMES;
      else
         to_write=dur_num_of_frames;
         
      if ((written = snd_pcm_writei (playback_handle, (void *)buf, to_write)) <=0) 
      {
         fprintf(stderr, "write to audio interface failed (%d/%d/%s)\n",
                 written, to_write, snd_strerror (written));
         return;
      }/* if */     
      dur_num_of_frames -= written;
      
   }/* while */
   
   
}/* putstat_talsa() */

/********************************************************************

NAME
	exit_talsa()

SYNOPSIS
        #include "mtx_alsa.h"
	exit_alsa();

DESCRIPTION
        close all soundcard connection
        
ARGUMENTS
	---

RETURN VALUE
        void
NOTES
       needs libasound2-dev (lib-alsa-sound)
SEE ALSO
       --- 
Date       Changes
------------------
25.02.2007 under construction

*********************************************************************/

void exit_talsa(void)
{
   snd_pcm_close (playback_handle);
   
}/* exit_talsa() */
