#include <sys/types.h>
#include <sys/stat.h>
#include <curses.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mrx.h"

#define FUNK_Y 3 /* height of function key window */
#define FUNK_W 16
#define CWWIN_TITLE "Receiving(Stop:F9)"
static char *FwwinTitle[5]={"Speed(F1+,F2-)","Tone(F3-,F4+)","Level(F5-,F6+)","",""};

/***********************************/

WINDOW *init_cwwin(WINDOW *mainwin, const char *s)
{
   WINDOW *cwwin1; /* frame of cwwin2      */
   WINDOW *cwwin2; /* cw output window     */
   int mx, my;     /* dimensions of screen */
   
   if(!mainwin)
   {
      fprintf(stderr,"%s,%d: arg error!\n",__FILE__, __LINE__);
      return NULL;
   }
   getmaxyx(mainwin,my,mx); /* macro ! */
   if(my<FUNK_Y-2 || mx<2)
   {
      fprintf(stderr,"%s,%d: height/width error!\n",__FILE__, __LINE__);
      return NULL;
   }
   cwwin1=subwin(mainwin, my-FUNK_Y, mx, 0, 0);
   if(!cwwin1)
   {
      fprintf(stderr,"%s,%d: subwin() error!\n",__FILE__, __LINE__);
      return NULL;
   }
   box(cwwin1,0,0);
   if(s)
   {
      mvaddstr(0,1,s);
   }
   wrefresh(cwwin1);
   cwwin2=subwin(mainwin, my-FUNK_Y-2, mx-2, 1, 1);
   if(!cwwin2)
   {
      fprintf(stderr,"%s,%d: subwin() error!\n",__FILE__, __LINE__);
      return NULL;
   }
   scrollok(cwwin2, TRUE);
   return cwwin2;

}/* init_cwwin() */

/***********************************/

WINDOW *init_funcwin(WINDOW *mainwin, int num, const char *s)
{
   WINDOW *win1; /* frame of win  */
   WINDOW *win2; /* output window */
   int mx, my;/* dimensions of screen */
   if(!mainwin)
   {
      fprintf(stderr,"%s,%d: arg error!\n",__FILE__, __LINE__);
      return NULL;
   }
   getmaxyx(mainwin,my,mx); /* macro ! */
   if(my<FUNK_Y-2 || mx<2)
   {
      fprintf(stderr,"%s,%d: height/width error!\n",__FILE__, __LINE__);
      return NULL;
   }
   win1=subwin(mainwin, FUNK_Y, FUNK_W, my-FUNK_Y, FUNK_W*num);
   if(!win1)
   {
      fprintf(stderr,"%s,%d: subwin() error!\n",__FILE__, __LINE__);
      return NULL;
   }
   box(win1,0,0);
   if(s)
   {
      mvaddstr(my-FUNK_Y, FUNK_W*num+1, s);
   }

   wrefresh(win1);
   win2=subwin(mainwin, FUNK_Y-2, FUNK_W-2, my-FUNK_Y+1, FUNK_W*num+1);
   if(!win2)
   {
      fprintf(stderr,"%s,%d: subwin() error!\n",__FILE__, __LINE__);
      return NULL;
   }
   scrollok(win2, TRUE);                           
   return win2;
}

/***********************************/

WINDOW *init_mainwin(void)
{
   WINDOW *mainwin;
   
   mainwin=initscr();
   cbreak();  /* no line buffering */
   noecho();  /* no char echo      */
   nodelay(mainwin,TRUE); /* non-blocking io   */
   nonl();
   intrflush(stdscr, FALSE);
   keypad(stdscr, TRUE);
   clear();
   return mainwin;
   
}/* init_mainwin() */

/***********************************/

void exit_mainwin(void)
{
/*   getch(); */
   endwin();
   
}/* exit_mainwin() */
/***********************************/

int setflags(int fdes, int flagchange)
{
   int flags, rc;
   
   flags=fcntl(fdes, F_GETFL);
   if(flags<0) 
   {
      fprintf(stderr,"%s,%d: fcntl():%s!\n", __FILE__, __LINE__, strerror(errno));
      return -1;
   }
   rc=fcntl(fdes, F_SETFL, flagchange^flags);
   if(rc<0) 
   {
      fprintf(stderr,"%s,%d: fcntl():%s!\n", __FILE__, __LINE__, strerror(errno));    
      return -2;
   }
   return 0;
   
}/* setflags() */

/***********************************/

int opencwpipe(pid_t *ppid)
{
   int fd[2];
   int pid;
   int rc;
   
   rc=pipe(fd);
   if(rc)
   {
      fprintf(stderr,"%s,%d: pipe(): %s!\n",__FILE__,__LINE__,strerror(errno));
      return -1;
   }
   
   pid=fork();
   if(pid<0)
   {
      fprintf(stderr,"%s,%d: fork(): %s!\n",__FILE__,__LINE__,strerror(errno));
      return -1;
   }
   else if(pid!=0) /* parent */
   {
      close(fd[1]); /* close write */
   }
   else /* child */
   {
      close(fd[0]); /* close read */
      close(STDOUT_FILENO);
      dup2(fd[1], STDOUT_FILENO); /* copy "content" of fd[1] to standard out */
      execlp("./mrx", "./mrx", "-c", NULL);
      fprintf(stderr, "%s,%d: execlp(): %s!\n",__FILE__,__LINE__,strerror(errno));
      exit(-1); /* child process only */
   }
   
   rc=setflags(fd[0], O_NONBLOCK);
   if(rc)
   {
      fprintf(stderr,"%s,%d: setflags(): %s!\n",__FILE__,__LINE__,strerror(errno));
      return -1;
   }
   *ppid=pid;
   return fd[0];
   
}/* opencwpipe() */

/***********************************/

void closecwpipe(int fdes, int pid)
{
   if(close(fdes))
   {
      fprintf(stderr,"%s,%d: close(%d): %s!\n",
      __FILE__, __LINE__, fdes, strerror(errno));
   }
   if(kill(pid, SIGTERM))
   {
      fprintf(stderr,"%s,%d: kill(%d,%d): %s!\n",
      __FILE__, __LINE__, pid, SIGTERM, strerror(errno));
   }
}/* closecwpipe() */

/***********************************/

int opencontrolpipe(const char *fifoname)
{
   int fdes;
   int rc;
   
   rc=mknod(fifoname, S_IFIFO|0600, 0);
   if(rc!=0&&errno!=EEXIST)
   {
      fprintf(stderr,"%s,%d: mknod(%s): %s!\n",
      __FILE__, __LINE__, fifoname, strerror(errno));
      return -1;
   }

   fdes=open(fifoname, O_WRONLY);
   if(fdes==-1)
   {
      fprintf(stderr,"%s,%d: open(%s): %s!\n",
      __FILE__, __LINE__, fifoname, strerror(errno));
      return -1;
   }

   return fdes;
   
}/* opencontrolpipe() */

/***********************************/

void closecontrolpipe(int fdes)
{
   if(close(fdes))
   {
      fprintf(stderr,"%s,%d: close(%d): %s!\n",
      __FILE__, __LINE__, fdes, strerror(errno));
   }
}/* closecontrolpipe() */

/***********************************/

void evaluate(const char *s, WINDOW *fwin[5])
{
   switch(s[0])
   {
   case 'S': wprintw(fwin[0],"\r   %s WPM  ", s+1); wrefresh(fwin[0]);break;
   case 'F': wprintw(fwin[1],"\r   %s Hz  ", s+1);  wrefresh(fwin[1]);break;
   case 'L': wprintw(fwin[2],"\r   %s  ", s+1);     wrefresh(fwin[2]);break;
   default: wprintw(fwin[3],"\r%s  ", s);           wrefresh(fwin[3]);break;
   }
   
}/* evaluate() */

/**********************************/

int main(void)
{
   pid_t cpid=0;
   int cwmode;
   WINDOW *cwwin;
   int end;
   int fdes;
   int fdes2;
   WINDOW *fwin[5];
   WINDOW *mainwin;
   int i;

   fdes=opencwpipe(&cpid);
   if(fdes==-1)
   {
      fprintf(stderr,"%s,%d: opencwpipe(): error!\n", __FILE__, __LINE__);
      return 1;
   }
   
   fdes2=opencontrolpipe(FIFONAME);
   if(fdes2==-1)
   {
       fprintf(stderr,"%s,%d: opencwpipe(): error!\n", __FILE__, __LINE__);
       return 2;
   }
   
   if((mainwin=init_mainwin())==NULL)
   {
      fprintf(stderr,"%s,%d: init_mainwin(): error!\n", __FILE__, __LINE__);
      exit_mainwin();
      return 3;
   }
   
   if((cwwin=init_cwwin(mainwin, CWWIN_TITLE))==NULL)
   {
      fprintf(stderr,"%s,%d: init_cwwin(): error!\n", __FILE__, __LINE__);
      exit_mainwin();
      return 4;
   }

   for(i=0; i<5; ++i)      
   {
      fwin[i]=init_funcwin(mainwin,i, FwwinTitle[i]);
      if(!fwin[i])
      {
         fprintf(stderr,"%s,%d: init_fwin(%d): error!\n", __FILE__, __LINE__,i);
         exit_mainwin();
         return 5;
      }
   }
   
   end=0;
   cwmode=1;
   while(!end)
   {
      int cnt;
      char s[2];
      char t[2];
      char reply[25];

      cnt=read(fdes, s, 1);
      if(cnt==1) 
      {
         switch(s[0])
         {
         case '[': cwmode=0; reply[0]=0;           break;
         case ']': cwmode=1; evaluate(reply,fwin); break;
         default: 
            if(cwmode)
               waddch(cwwin,s[0]);
            else
               if(strlen(reply)<sizeof(reply)-1) strncat(reply, s, 1);
         }
      }

      i=getch();
      switch(i)
      {
      case ERR: break;
      case KEY_F(1): t[0]='S'; write(fdes2,t,1); break;
      case KEY_F(2): t[0]='s'; write(fdes2,t,1); break;
      case KEY_F(3): t[0]='F'; write(fdes2,t,1); break;
      case KEY_F(4): t[0]='f'; write(fdes2,t,1); break;
      case KEY_F(5): t[0]='L'; write(fdes2,t,1); break;
      case KEY_F(6): t[0]='l'; write(fdes2,t,1); break;
      case KEY_F(8): waddch(cwwin, '8');        break;
      case KEY_F(9): end=1;                     break;
      default: break;
      }
      wrefresh(cwwin);
   }
   closecwpipe(fdes, cpid);
   closecontrolpipe(fdes2);
   exit_mainwin();
   return 0;
   
}/* main() */
/**********************************/
